from setuptools import find_packages, setup
setup(
    name='facefeaturelib',
    packages=find_packages(include=['facefeaturelib']),
    version='0.1.0',
    description='image feature library',
    author='hoangvd3',
    license='MIT',
    install_requires=[],
)